/*
 *
 *	Utility for reporting quotas
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <errno.h>

#include <asm/byteorder.h>

#include "common.h"
#include "quotasys.h"
#include "quota_io.h"

#define FL_USER 1
#define FL_GROUP 2
#define FL_VERBOSE 4
#define FL_ALL 8
#define FL_STAT 16

#define set_bit(bmp, ind) ((bmp)[(ind) >> 3] |= (1 << ((ind) & 7)))
#define get_bit(bmp, ind) ((bmp)[(ind) >> 3] & (1 << ((ind) & 7)))

int flags;
char **mnt;
int mntcnt;
char *bitmap;

void parse_options(int argcnt, char **argstr)
{
	int ret;
	char *slash = strrchr(argstr[0], '/');

	if (!slash)
		slash = argstr[0];
	else
		slash++;

	while ((ret = getopt(argcnt, argstr, "Vavughs")) != EOF) {
		switch(ret) {
			case '?':
			case 'h':
usage:
				printf("Utility for reporting quotas.\nUsage:\n%s [-vug] (-a | mntpoint)\n", slash);
				printf("Bugs to %s\n", MY_EMAIL);
				exit(1);
			case 'V':
				version();
				exit(0);
			case 'u':
				flags |= FL_USER;
				break;
			case 'g':
				flags |= FL_GROUP;
				break;
			case 'v':
				flags |= FL_VERBOSE;
				break;
			case 'a':
				flags |= FL_ALL;
				break;
			case 's':
				flags |= FL_STAT;
				break;
		}
	}

	if ((flags & FL_ALL && optind != argcnt) || (!(flags & FL_ALL) && optind == argcnt)) {
		fputs("Bad number of arguments.\n", stderr);
		goto usage;
	}
	if (!(flags & (FL_USER | FL_GROUP)))
		flags |= FL_USER;
	if (!(flags & FL_ALL)) {
		mnt = argstr + optind;
		mntcnt = argcnt - optind;
	}
}

char overlim(uint usage, uint softlim, uint hardlim)
{
	if ((usage > softlim && softlim) || (usage > hardlim && hardlim))
		return '+';
	return '-';
}

void print(qid_t id, short type, struct mem_dqblk *entry)
{
	char name[MAXNAMELEN];
	char time[MAXTIMELEN];

	if (!entry->dqb_curspace && !entry->dqb_curinodes && !(flags & FL_VERBOSE))
		return;
	if (type == QS_USRQUOTA)
		uid2name(id, name);
	else
		gid2group(id, name);
	difftime2str(entry->dqb_btime, time);
	printf("%-10s%c%c%8d%8d%8d%7s", name, overlim(qb2kb(toqb(entry->dqb_curspace)), qb2kb(entry->dqb_bsoftlimit), qb2kb(entry->dqb_bhardlimit)),
	  overlim(entry->dqb_curinodes, entry->dqb_isoftlimit, entry->dqb_ihardlimit), (int)qb2kb(toqb(entry->dqb_curspace)), qb2kb(entry->dqb_bsoftlimit), qb2kb(entry->dqb_bhardlimit), time);
	difftime2str(entry->dqb_itime, time);
	printf("%8d%6d%6d%7s\n", entry->dqb_curinodes, entry->dqb_isoftlimit, entry->dqb_ihardlimit, time);
}

int report_block(uint blk, short type)
{
	dqbuf_t buf = getdqbuf();
	struct disk_dqdbheader *dh;
	struct disk_dqblk *ddata;
	struct mem_dqblk mdata;
	int entries, i;

	set_bit(bitmap, blk);
	read_blk(blk, buf);
	dh = (struct disk_dqdbheader *)buf;
	ddata = (struct disk_dqblk *)(((char *)buf) + sizeof(struct disk_dqdbheader));
	entries = __le16_to_cpu(dh->dqdh_entries);
	for (i = 0; i < DQSTRINBLK; i++)
		if (!empty_dquot(ddata+i)) {
			disk2memdqb(&mdata, ddata + i);
			print(__le32_to_cpu(ddata[i].dqb_id), type, &mdata);
		}
	freedqbuf(buf);
	return entries;
}

int report_tree(uint blk, int depth, short type)
{
	int entries = 0, i;
	dqbuf_t buf = getdqbuf();
	__u32 *ref = (__u32 *)buf;

	read_blk(blk, buf);
	if (depth == DQTREEDEPTH-1) {
		for (i = 0; i < DQBLKSIZE >> 2; i++) {
			blk = __le32_to_cpu(ref[i]);
			if (blk && !get_bit(bitmap, blk))
				entries += report_block(__le32_to_cpu(ref[i]), type);
		}
	}
	else {
		for (i = 0; i < DQBLKSIZE >> 2; i++)
			if (__le32_to_cpu(ref[i]))
				entries += report_tree(__le32_to_cpu(ref[i]), depth+1, type);
	}
	freedqbuf(buf);
	return entries;
}

uint find_set_bits(char *bmp, int blocks)
{
	uint i, used = 0;

	for (i = 0; i < blocks; i++)
		if (get_bit(bmp, i))
			used++;
	return used;
}

void report_it(char *mnt, int type, char *dev)
{
	int entries = 0;
	char bgbuf[MAXTIMELEN], igbuf[MAXTIMELEN];

	init_io(mnt, dev, type);
	if (QIO_ENABLED && quotactl(QCMD(Q_SYNC, user2kernquota(type)), dev, 0, NULL) < 0)
		fprintf(stderr, "Warning: Can't sync quotas (%s). Results are not reliable...\n", strerror(errno));
	printf("*** Report for %s quotas on device %s (%s)\n", type2name(type), dev, mnt);
	time2str(info.dqi_bgrace, bgbuf, TF_ROUND);
	time2str(info.dqi_igrace, igbuf, TF_ROUND);
	printf("Block grace time: %s Inode grace time: %s\n", bgbuf, igbuf);
	printf("                        Block limits               File limits\n");
	printf("User            used    soft    hard  grace    used  soft  hard  grace\n");
	
	bitmap = smalloc((info.dqi_blocks+7)>>3);
	memset(bitmap, 0, (info.dqi_blocks+7)>>3);
	entries = report_tree(1, 0, type);
	if (flags & FL_STAT) {
		uint blk = find_set_bits(bitmap, info.dqi_blocks);
		printf("Statistics:\nBlocks: %u\nEntries: %d\nUsed average: %f\n", info.dqi_blocks, entries, ((float)entries)/blk);
	}
	putchar('\n');
	free(bitmap);
	free_io();
}

void report(int type)
{
	char devname[MAXDEVLEN];
	char filename[MAXFILENAMELEN];

	init_mount(mnt, mntcnt);
	while (next_mount(type, devname, filename, NULL))
		report_it(filename, type, devname);
	free_mount();
}

int main(int argcnt, char **argstr)
{
	parse_options(argcnt, argstr);
	if (flags & FL_USER)
		report(QS_USRQUOTA);
	if (flags & FL_GROUP)
		report(QS_GRPQUOTA);
	return 0;
}

